"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.DogeTransaction = void 0;
const constants_1 = require("../../utils/constants");
const errors_1 = require("../../utils/errors");
const utils_1 = require("../../utils/utils");
const TransactionBase_1 = require("../TransactionBase");
const DogeAddress_1 = require("../addressObjects/DogeAddress");
const UtxoTransaction_1 = require("./UtxoTransaction");
class DogeTransaction extends UtxoTransaction_1.UtxoTransaction {
    constructor(data, additionalData) {
        super(data, additionalData);
        this.data.vout.forEach((vout) => {
            this.processOutput(vout);
        });
        this.synchronizeAdditionalData();
        this.assertAdditionalData();
    }
    isValidPkscript(index) {
        const vout = this.extractVoutAt(index);
        if (!vout.scriptPubKey.address)
            return true; //OP_RETURN
        const address = new DogeAddress_1.DogeAddress(vout.scriptPubKey.address);
        return address.addressToPkscript() == vout.scriptPubKey.hex;
    }
    get currencyName() {
        return constants_1.DOGE_NATIVE_TOKEN_NAME;
    }
    async paymentSummary({ transactionGetter, inUtxo, outUtxo }) {
        try {
            this.assertValidVinIndex(inUtxo);
        }
        catch (e) {
            return { status: TransactionBase_1.PaymentSummaryStatus.InvalidInUtxo };
        }
        try {
            this.assertValidVoutIndex(outUtxo);
        }
        catch (e) {
            return { status: TransactionBase_1.PaymentSummaryStatus.InvalidOutUtxo };
        }
        // if (transactionGetter === undefined) {
        //     return { status: PaymentSummaryStatus.NoTransactionGetterProvided };
        // }
        const a = transactionGetter;
        //console.log(a);
        await this.vinVoutAt(inUtxo, transactionGetter);
        if (this.type === "coinbase") {
            return { status: TransactionBase_1.PaymentSummaryStatus.Coinbase };
        }
        const spentAmount = this.spentAmounts[inUtxo];
        if (!spentAmount.address) {
            return { status: TransactionBase_1.PaymentSummaryStatus.NoSpentAmountAddress };
        }
        const receiveAmount = this.receivedAmounts[outUtxo];
        if (!receiveAmount.address) {
            return { status: TransactionBase_1.PaymentSummaryStatus.NoReceiveAmountAddress };
        }
        // Extract addresses from input and output fields
        const sourceAddress = spentAmount.address;
        const receivingAddress = receiveAmount.address;
        // We will update this once we iterate over inputs and outputs if we have full transaction
        let oneToOne = true;
        const isFull = this.type === "full_payment";
        if (isFull) {
            let inFunds = (0, utils_1.toBN)(0);
            let returnFunds = (0, utils_1.toBN)(0);
            let outFunds = (0, utils_1.toBN)(0);
            let inFundsOfReceivingAddress = (0, utils_1.toBN)(0);
            for (const vinAmount of this.spentAmounts) {
                if (sourceAddress && vinAmount.address === sourceAddress) {
                    inFunds = inFunds.add(vinAmount.amount);
                }
                if (receivingAddress && vinAmount.address === receivingAddress) {
                    inFundsOfReceivingAddress = inFundsOfReceivingAddress.add(vinAmount.amount);
                }
                if (oneToOne && vinAmount.address != sourceAddress && vinAmount.address != receivingAddress) {
                    oneToOne = false;
                }
            }
            for (const voutAmount of this.receivedAmounts) {
                if (sourceAddress && voutAmount.address === sourceAddress) {
                    returnFunds = returnFunds.add(voutAmount.amount);
                }
                if (receivingAddress && voutAmount.address === receivingAddress) {
                    outFunds = outFunds.add(voutAmount.amount);
                }
                // Outputs without address do not break one-to-one condition
                if (oneToOne && !voutAmount.address && voutAmount.amount.gt((0, utils_1.toBN)(0))) {
                    oneToOne = false;
                }
                if (oneToOne && voutAmount.address && voutAmount.address != sourceAddress && voutAmount.address != receivingAddress) {
                    oneToOne = false;
                }
            }
            return {
                status: TransactionBase_1.PaymentSummaryStatus.Success,
                response: {
                    blockTimestamp: this.unixTimestamp,
                    transactionHash: this.stdTxid,
                    sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                    sourceAddress,
                    receivingAddress,
                    receivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                    spentAmount: inFunds.sub(returnFunds),
                    receivedAmount: outFunds.sub(inFundsOfReceivingAddress),
                    transactionStatus: this.successStatus,
                    paymentReference: this.stdPaymentReference,
                    // Intended and actual amounts are the same for utxo transactions
                    intendedSourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                    intendedSourceAddress: sourceAddress,
                    intendedSourceAmount: inFunds.sub(returnFunds),
                    intendedReceivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                    intendedReceivingAddress: receivingAddress,
                    intendedReceivingAmount: outFunds.sub(inFundsOfReceivingAddress),
                    oneToOne,
                    isFull,
                },
            };
        }
        else {
            // Since we don't have all inputs "decoded" we can't be sure that transaction is one-to-one
            oneToOne = false;
            const spentAmount = sourceAddress && inUtxo != null ? this.spentAmounts[inUtxo].amount : (0, utils_1.toBN)(0);
            const receivedAmount = receivingAddress && outUtxo != null ? this.receivedAmounts[outUtxo].amount : (0, utils_1.toBN)(0);
            return {
                status: this.isValidPkscript(outUtxo) ? TransactionBase_1.PaymentSummaryStatus.Success : TransactionBase_1.PaymentSummaryStatus.InvalidPkscript,
                response: {
                    blockTimestamp: this.unixTimestamp,
                    transactionHash: this.stdTxid,
                    sourceAddress,
                    sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                    receivingAddress,
                    receivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                    spentAmount,
                    receivedAmount,
                    paymentReference: this.stdPaymentReference,
                    transactionStatus: this.successStatus,
                    // Intended and actual amounts are the same for utxo transactions
                    intendedSourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                    intendedSourceAddress: sourceAddress,
                    intendedSourceAmount: spentAmount,
                    intendedReceivingAddressHash: (0, utils_1.standardAddressHash)(receivingAddress),
                    intendedReceivingAddress: receivingAddress,
                    intendedReceivingAmount: receivedAmount,
                    oneToOne,
                    isFull,
                },
            };
        }
    }
    async balanceDecreasingSummary({ sourceAddressIndicator, transactionGetter, }) {
        var _a, _b;
        // We expect sourceAddressIndicator to be utxo vin index (as hex string)
        if (!(0, utils_1.isValidBytes32Hex)(sourceAddressIndicator)) {
            return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NotValidSourceAddressFormat };
        }
        const vinIndex = parseInt(sourceAddressIndicator, 16);
        if (isNaN(vinIndex)) {
            return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NotValidSourceAddressFormat };
        }
        try {
            this.assertValidVinIndex(vinIndex);
        }
        catch (e) {
            return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.InvalidInUtxo };
        }
        const transactionType = this.type;
        console.log(transactionType);
        switch (transactionType) {
            case "coinbase": {
                return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.Coinbase };
            }
            case "full_payment": {
                const spentAmounts = this.spentAmounts;
                const spentAmount = spentAmounts[vinIndex];
                const sourceAddress = spentAmount.address;
                if (sourceAddress) {
                    let inFunds = (0, utils_1.toBN)(0);
                    let returnFunds = (0, utils_1.toBN)(0);
                    for (const vinAmount of this.spentAmounts) {
                        if (sourceAddress && vinAmount.address === sourceAddress) {
                            inFunds = inFunds.add(vinAmount.amount);
                        }
                    }
                    for (const voutAmount of this.receivedAmounts) {
                        if (sourceAddress && voutAmount.address === sourceAddress) {
                            returnFunds = returnFunds.add(voutAmount.amount);
                        }
                    }
                    return {
                        status: TransactionBase_1.BalanceDecreasingSummaryStatus.Success,
                        response: {
                            blockTimestamp: this.unixTimestamp,
                            transactionHash: this.stdTxid,
                            sourceAddressIndicator: sourceAddressIndicator,
                            sourceAddressHash: (0, utils_1.standardAddressHash)(sourceAddress),
                            sourceAddress: sourceAddress,
                            spentAmount: inFunds.sub(returnFunds),
                            transactionStatus: this.successStatus,
                            paymentReference: this.stdPaymentReference,
                            isFull: true,
                        },
                    };
                }
                else {
                    // The input has no address, the type is wrongly extracted (Should not happen)
                    return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.InvalidTransactionDataObject };
                }
                return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.Success }; // TODO: implement
            }
            case "partial_payment": {
                if (!transactionGetter) {
                    throw new Error("Transaction getter not provided");
                }
                const vinVout = await this.extractVinVoutAt(vinIndex, transactionGetter);
                if ((_b = (_a = vinVout === null || vinVout === void 0 ? void 0 : vinVout.vinvout) === null || _a === void 0 ? void 0 : _a.scriptPubKey) === null || _b === void 0 ? void 0 : _b.address) {
                    return {
                        status: TransactionBase_1.BalanceDecreasingSummaryStatus.Success,
                        response: {
                            blockTimestamp: this.unixTimestamp,
                            transactionHash: this.stdTxid,
                            sourceAddressIndicator: sourceAddressIndicator,
                            sourceAddressHash: (0, utils_1.standardAddressHash)(vinVout.vinvout.scriptPubKey.address),
                            sourceAddress: vinVout.vinvout.scriptPubKey.address,
                            spentAmount: (0, utils_1.toBN)(Math.round((vinVout.vinvout.value || 0) * constants_1.BTC_MDU).toFixed(0)),
                            transactionStatus: this.successStatus,
                            paymentReference: this.stdPaymentReference,
                            isFull: false,
                        },
                    };
                }
                return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NoSourceAddress };
            }
            default:
                // exhaustive switch guard: if a compile time error appears here, you have forgotten one of the cases
                // eslint-disable-next-line @typescript-eslint/no-empty-function, @typescript-eslint/no-unused-vars
                (_) => { };
                // We didn't find the address we are looking for
                return { status: TransactionBase_1.BalanceDecreasingSummaryStatus.NoSourceAddress };
        }
    }
    async makeFull(transactionGetter) {
        await this.makeFullPayment(transactionGetter);
    }
    ///////////////////////////////
    //// Utxo specific methods ////
    ///////////////////////////////
    /**
     * Asserts whether vinvouts Mapper is full mapper and only has unique and matching indices
     * This mapper is required to make full transactions on DOGE chain
     */
    assertAdditionalData() {
        var _a;
        if (this.additionalData) {
            if (((_a = this.additionalData.vinouts) === null || _a === void 0 ? void 0 : _a.length) !== this.data.vin.length) {
                throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Bad format, Additional data vinvouts and data vin length mismatch"));
            }
            this.additionalData.vinouts.forEach((vinvout, ind) => {
                if (vinvout && vinvout.index !== ind) {
                    throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, Error("Additional data corrupted: indices mismatch"));
                }
            });
        }
    }
    /**
     * Checks that additional data mapper for making full transactions is correctly formatted
     * @returns boolean indicator if the object is correctly formatted
     * This mapper is required to make full transactions on DOGE chain
     */
    isValidAdditionalData() {
        var _a;
        if (this.additionalData) {
            if (((_a = this.additionalData.vinouts) === null || _a === void 0 ? void 0 : _a.length) !== this.data.vin.length) {
                return false;
            }
            this.additionalData.vinouts.forEach((vinvout, ind) => {
                if (vinvout && vinvout.index !== ind) {
                    return false;
                }
            });
            return true;
        }
        return false;
    }
    synchronizeAdditionalData() {
        const tempAdditionalData = this.additionalData;
        this.additionalData = {
            vinouts: new Array(this.data.vin.length).fill(undefined),
        };
        if (tempAdditionalData === null || tempAdditionalData === void 0 ? void 0 : tempAdditionalData.vinouts) {
            for (const vinvout of tempAdditionalData.vinouts) {
                if (vinvout) {
                    if (vinvout.index < 0 || vinvout.index >= this.data.vin.length || !this.additionalData || !this.additionalData.vinouts) {
                        throw new errors_1.mccError(errors_1.mccErrorCode.InvalidParameter, new Error("vinvout wrong index out of range"));
                    }
                    this.additionalData.vinouts[vinvout.index] = vinvout;
                    this.processOutput(vinvout.vinvout);
                }
            }
        }
    }
    /**
     *
     * @param vinIndex vin index
     * @returns vin on vin index
     */
    extractVinAt(vinIndex) {
        this.assertValidVinIndex(vinIndex);
        return this.data.vin[vinIndex];
    }
    /**
     * Gets the vout corresponding to vin in the given vin index. If the vout is not fetched, the
     * input transaction is read and the corresponding vout is obtained
     * @param vinIndex vin index
     * @param client mcc client for fetching input the transactions
     * @returns input vout of the relevant utxo of the input transaction on the vin on the index `vinIndex`
     */
    async extractVinVoutAt(vinIndex, transactionGetter) {
        const vinObject = this.extractVinAt(vinIndex);
        if (!vinObject.txid) {
            return {
                index: vinIndex,
                vinvout: undefined,
            };
        }
        const connectedTrans = await transactionGetter(vinObject.txid);
        if (connectedTrans === null) {
            return {
                index: vinIndex,
                vinvout: undefined,
            };
        }
        return {
            index: vinIndex,
            vinvout: connectedTrans.extractVoutAt(vinObject.vout || 0),
        };
    }
    /**
     * Gets the vout corresponding to vin in the given vin index. If it is not stored in the additionalData,
     * it's fetched and updated using the client.
     * @param vinIndex vin index
     * @param client mcc client to fetch transaction data
     * @returns input vout of the relevant utxo of the input transaction on the vin on the index `vinIndex`
     */
    async vinVoutAt(vinIndex, transactionGetter) {
        // note: vinouts list is always initialized
        if (!this.additionalData) {
            this.additionalData = {
                vinouts: new Array(this.data.vin.length).fill(undefined),
            };
        }
        if (!this.additionalData || !this.additionalData.vinouts) {
            throw new errors_1.mccError(errors_1.mccErrorCode.NotImplemented, new Error(`Can not happen`));
        }
        let vinVout = this.additionalData.vinouts[vinIndex];
        if (vinVout) {
            return vinVout;
        }
        if (!transactionGetter) {
            throw new Error("Transaction getter not provided");
        }
        vinVout = await this.extractVinVoutAt(vinIndex, transactionGetter);
        this.additionalData.vinouts[vinIndex] = vinVout;
        return vinVout;
    }
    /**
     * Checks if the input on index `vinIndex` is fetched.
     * @param vinIndex vin index
     * @returns true if input on the index `vinIndex` is already fetched.
     */
    isSyncedVinIndex(vinIndex) {
        this.assertValidVinIndex(vinIndex);
        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
        return !!this.additionalData.vinouts[vinIndex];
    }
    /**
     * Fetches all the vin data from input transactions and makes the payment full_payment.
     * @param client mcc client to use for fetching input data
     */
    async makeFullPayment(transactionGetter) {
        const promises = [];
        this.synchronizeAdditionalData();
        for (let i = 0; i < this.data.vin.length; i++) {
            promises.push(this.vinVoutAt(i, transactionGetter));
        }
        await Promise.all(promises);
    }
    /**
     * Doge has additional brackets around addresses in out tx
     * @param vout
     * @returns
     */
    processOutput(vout) {
        if (!vout) {
            return;
        }
        if (vout.scriptPubKey.address) {
            return;
        }
        if (vout.scriptPubKey.addresses && vout.scriptPubKey.addresses.length === 1) {
            vout.scriptPubKey.address = vout.scriptPubKey.addresses[0];
        }
        // otherwise `address` stays undefined
    }
}
exports.DogeTransaction = DogeTransaction;
//# sourceMappingURL=DogeTransaction.js.map